require 'nixadm/util'
require 'nixadm/backup'

module NixAdm
module PostgreSQL

module Replication

class Node

  include NixAdm::Util

  def initialize(host, config_file=nil)
    @host = host
    @db   = connectDb(host, config_file)

    if @db.nil?
      raise "Connection failed to #{host}"
    end

    logSystemInit(config_file)
    @logfields[:system] = 'psql'
  end

  def status()
    return @db.exec('select * from pg_stat_replication')
  end

end # class Node

# Primary database instance
class Primary < Node

  include NixAdm::Util

  attr_reader :db

  def initialize(host, config_file=nil)
    super
  end

  def currentLsm()
    return @db.exec('select pg_current_wal_lsn()')[0]
  end

end # class Master

# Hot standby instance
class Secondary < Node

  include NixAdm::Util

  attr_reader :primary

  def initialize(primary, host, config_file=nil)
    @primary = primary

    super(host, config_file)
  end

  def currentLsm()
    return @db.exec('select pg_last_wal_receive_lsn()')[0]
  end

end # class Replica

end # module Replication

# This class exports an entire PostgreSQL database cluster to $backup_dir/db,
# including the relevant global settings.
#
# Example:
#
#   export = Export.new('db', { :debug => true })
#   export.run()

class Export < NixAdm::Backup

  attr_accessor :filesystems, :targets

  def initialize(host, options = {})
    super host, options: options

    @ssh_opts = '-o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no'
  end

  def log(msg, code=0)
    $stderr.puts msg
  end

  #------------------------------------------------------------------------------
  # Main backup function
  #------------------------------------------------------------------------------

  def run()
    begin
      backup_databases()
    rescue Exception => ex
      log "Backup failed: #{ex.to_s}\n #{ex.backtrace}"
    end
  end

  #------------------------------------------------------------------------------
  # Individual backup procedures
  #------------------------------------------------------------------------------

  # This does a pg_dump of all databases in plaintext format. It gzips them all.
  def backup_databases()
    cd backup_dir('db')

    # List of databases on server
    databases = listDatabases()

    # List of databases to exclude
    exclude =
    {
      'template0' => true,
      'template1' => true
    }

    # Backup individual databases
    databases.each do |db|
      next if exclude.has_key?(db)

      begin
        log "Dumping: #{db}"
        store @host, "pg_dump -p 5433 #{db}", "#{db}.db"
      rescue
        puts $!
        log "Failed #{@host} #{db}: #{$?.to_s}", 1
        next
      end
    end

    # Dump the globals
    globals_file = 'globals.sql'
    log "Dumping globals to #{globals_file}"
    store @host, 'pg_dumpall -g -p 5433', globals_file
  end

  protected

  def listDatabases()
    command = %Q{ psql -At -p 5433 postgres -c 'select datname from pg_database' }
    pipeline = resolveCommand(command)

    dbs = []
    exec([pipeline]) do |line|
      dbs << line.strip
    end

    return dbs
  end

end # class Export

# This class imports everything exported from the Export class. Before
# importing, you need to make sure you initialize your database with proper
# encoding and collation. To do so, switch user to postgres. Then run:
#
#  /usr/local/bin/initdb --pgdata=/var/lib/pgsql9/data/ -E 'UTF-8' \
#    --lc-collate='en_US.UTF-8' --lc-ctype='en_US.UTF-8';
#
#  /usr/local/bin/createuser -sl root
#
# Then start Postgres. Now you can import databases

class Import < NixAdm::Backup

  def initialize(host, options = {})
    super host, options: options

    @ssh_opts = '-o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no'
  end

  #------------------------------------------------------------------------------
  # Main backup function
  #------------------------------------------------------------------------------

  def run()
    begin
      import_databases()
    rescue Exception => ex
      log "Import failed: #{ex.to_s}\n #{ex.backtrace}"
    end
  end

  def import_databases()
    dbs = Dir.entries($backup_dir)

    # Restore globals first
    system "psql -p 5433 -d postgres -f #{$backup_dir}/globals.sql"

    dbs.each do |file|
      next if File.extname(file) != '.db'

      db = file.split('.')[0]
      puts "Loading #{db}"

      sql   = %q{ select count(*) from pg_database where datname='dba' }
      count = %x{ psql -p 5433 postgres -At -c "#{sql}" }
      count.strip!

      if count == '1'
        %x{ psql -p 5433 postgres -c 'drop database #{db}' }
      end

      %x{ psql -p 5433 postgres -c 'create database #{db}' }
      %x{ psql -p 5433 -d #{db} -f #{$backup_dir}/#{file}  }
      %x{ psql -p 5433 #{db}    -c 'analyze' }
    end
  end

end # class Import

end # module PostgreSQL
end # module NixAdm
