require 'pg'
require 'rbconfig'
require 'securerandom'
require 'yaml'
require 'socket'

require 'nixadm/version'

#-------------------------------------------------------------------------------
# Global parameters
#-------------------------------------------------------------------------------

# Configuration file. Config file is YAML.

$nixadm_config_file = 'nixadm.conf'
#-------------------------------------------------------------------------------

class String

  def sqlEscape()
    return self.gsub("'", "''").gsub("\\", "\\\\")
  end

end

# This extends PG::Result making it more "rubyish"
module PG

class Result

  # Takes a row and column and return associated field. +r+ is the row
  # ordinal. +c+ can be either a string (column name) or an integer
  # (column ordinal).
  def [](r, c)
    idx = c

    if c.class == String
      idx = self.fnumber(c)
    end

    return self.getvalue(r, idx)
  end

  # This is meant to be used with +each()+. It takes a column and
  # returns the associated field value. The current row is the active
  # row in the +each()+ block. +c+ can be a String (column name) or an
  # integer (column ordinal).
  def [](c)
    idx = c

    if c.class == String
      idx = self.fnumber(c)
    end

    return self.getvalue(@row || 0, idx)
  end

end # class PG
end # module Result

module NixAdm

# Optional mixin for general utility functions

module Util

  # Run shell command in bash with pipefail option for full error detection on
  # pipelines.
  def bash(cmd)
    cmd = "set -o pipefail && #{cmd}"
    system("bash -c #{cmd.shellescape}")
    cmd_status = $?
    if cmd_status != 0
      raise "'#{cmd}' execution failed (exit code: #{cmd_status})"
    end
  end

  # Logging to PostgreSQL
  #
  # Connect to a database. Uses configuration file. Config file is YAML and
  # logger uses the following format for database connection:
  #
  # ---
  # databases:
  #   logger:
  #     host: db
  #     db: nixadm
  #     user: bob
  #     password: bobspassword
  #
  # @param db_name The name of the entry in the configuration file for the
  #   database connection paramaters.
  #
  # @return Database connection if successful, nil otherwise

  def logSystemInit(config_file = nil)
    @logdb = connectDb('logger', config_file)

    @logfields =
    {
      :host     =>  Socket.gethostname,
      :system   => 'backup',
      :module   => self.class.name,
      :function => 'run'
    }
  end

  # Call when complete backup is done. Disconnects from database

  def logSystemfinalize()
    @logdb.disconnect()
  end

  # Connect to a database. Uses configuration file. Config file is YAML and has
  # following format for databases:
  #
  # ---
  # databases:
  #   logger:
  #     host: db
  #     db: nixadm
  #     user: bob
  #     password: bobspassword
  #
  # @param db_name The name of the entry in the configuration file for the
  #   database connection paramaters.
  #
  # @return Database connection if successful, nil otherwise

  def connectDb(db_name, config_file = nil)

    if config_file.nil?
      config_file = $nixadm_config_file

      # Account for BSD systems that user /usr/local/etc
      prefix = RbConfig::CONFIG['prefix']
      if prefix == '/usr/local'
        config_file = prefix + $nixadm_config_file
      end
    end

    @config = YAML::load_file(config_file)

    params   = @config['databases'][db_name]

    login    = params['user']
    password = params['password']
    db       = params['db']
    host     = params['host']
    port     = params['port'] || '5432'

    return PG::Connection.new(host, port, '', '', db, login, password)
  end

  # Called before starting a backup job. Sets service function name based on
  # called method and logs a start message.

  def jobStart(host=nil)

    if not host.nil?
      @logfields[:function] = "#{caller_locations()[0].label} #{host} "
    else
      @logfields[:function] = "#{caller_locations()[0].label}"
    end

    @job = SecureRandom.uuid()

    log('Started', 0)
  end

  # Call when job has completed. Logs a completed message and clears the service
  # function name.

  def jobFinish()
    log('Completed', 0)
    @logfields[:function] = ''
  end

  # Logs a message. Uses service name defined in jobStart().
  #
  # @param msg A log message describing a completed operation
  # @param code The status code of the operation in the log message

  def log(message, code = 0)
    fields = {}.merge(@logfields)

    fields['job']  = @job
    fields['code'] = code

    if not message.nil?
      fields['message'] = message.sqlEscape()
    end

    keys   = []
    values = []

    fields.each do |k,v|
      keys   << k.to_s
      values << v
    end

    sql = %Q{insert into sys.log (#{keys.join(",")}) values ('#{values.join("','")}')}
    @logdb.exec(sql)

    $stderr.puts message
  end

end # module Util

end # module NixAdm
